#ifndef CORE_VECTOR_H
#define CORE_VECTOR_H

#include "core/Types.h"

#define CORE_VECTOR_OP2(name) name *r, const name *a
#define CORE_VECTOR_OP3(name) CORE_VECTOR_OP2(name), const name* b

#define CORE_DEFINE_VECTOR(N, name, sname, type, factor)                       \
    typedef struct {                                                           \
        type data[N];                                                          \
    } name;                                                                    \
    name* coreAddSet##sname(CORE_VECTOR_OP2(name));                            \
    name* coreAdd##sname(CORE_VECTOR_OP3(name));                               \
    name* coreSubSet##sname(CORE_VECTOR_OP2(name));                            \
    name* coreSub##sname(CORE_VECTOR_OP3(name));                               \
    name* coreMulSet##sname(CORE_VECTOR_OP2(name));                            \
    name* coreMul##sname(CORE_VECTOR_OP3(name));                               \
    name* coreDivSet##sname(CORE_VECTOR_OP2(name));                            \
    name* coreDiv##sname(CORE_VECTOR_OP3(name));                               \
    name* coreMulSet##sname##factor(name* r, type f);                          \
    name* coreMul##sname##factor(name* r, const name* a, type f);              \
    name* coreDivSet##sname##factor(name* r, type f);                          \
    name* coreDiv##sname##factor(name* r, const name* a, type f);              \
    name* coreInvertSet##sname(name* r);                                       \
    name* coreInvert##sname(name* r, const name* a);                           \
    type coreDot##sname(const name* a, const name* b);                         \
    type coreSquareLength##sname(const name* a);                               \
    type coreLength##sname(const name* a);                                     \
    name* coreNormalize##sname(name* r);                                       \
    void coreToString##sname(const name* a, char* buffer, size_t n);

CORE_DEFINE_VECTOR(2, CoreVector2, V2, float, F)
CORE_DEFINE_VECTOR(3, CoreVector3, V3, float, F)
CORE_DEFINE_VECTOR(4, CoreVector4, V4, float, F)
CORE_DEFINE_VECTOR(2, CoreIntVector2, IV2, int, I)
CORE_DEFINE_VECTOR(3, CoreIntVector3, IV3, int, I)
CORE_DEFINE_VECTOR(4, CoreIntVector4, IV4, int, I)

CoreVector3* coreAngles(CoreVector3* r, float lengthAngle, float widthAngle);
CoreVector3* coreCross(CORE_VECTOR_OP3(CoreVector3));

#define CORE_DEFINE_VECTOR_CONVERSION(a, b, name)                              \
    a* coreConvert##name(a* r, const b* c)

CORE_DEFINE_VECTOR_CONVERSION(CoreVector2, CoreIntVector2, IV2);
CORE_DEFINE_VECTOR_CONVERSION(CoreVector3, CoreIntVector3, IV3);
CORE_DEFINE_VECTOR_CONVERSION(CoreVector4, CoreIntVector4, IV4);
CORE_DEFINE_VECTOR_CONVERSION(CoreIntVector2, CoreVector2, V2);
CORE_DEFINE_VECTOR_CONVERSION(CoreIntVector3, CoreVector3, V3);
CORE_DEFINE_VECTOR_CONVERSION(CoreIntVector4, CoreVector4, V4);

#endif
