#include "core/Vector.h"

#include <math.h>
#include <stdio.h>

#include "core/Utility.h"

#define V2 CoreVector2
#define V3 CoreVector3
#define V4 CoreVector4

V3* coreAngles(V3* r, float lengthAngle, float widthAngle) {
    lengthAngle = coreDegreeToRadian(lengthAngle);
    widthAngle = coreDegreeToRadian(widthAngle);

    float sWidth = sinf(widthAngle);
    float cWidth = cosf(widthAngle);
    float sLength = sinf(lengthAngle);
    float cLength = cosf(lengthAngle);

    r->data[0] = cWidth * cLength;
    r->data[1] = sWidth;
    r->data[2] = -sLength * cWidth;
    return r;
}

V3* coreCross(V3* r, const V3* a, const V3* b) {
    r->data[0] = a->data[1] * b->data[2] - a->data[2] * b->data[1];
    r->data[1] = a->data[2] * b->data[0] - a->data[0] * b->data[2];
    r->data[2] = a->data[0] * b->data[1] - a->data[1] * b->data[0];
    return r;
}

#define VECTOR_IMPL(T, N, SN, F, CN)                                           \
    T* coreAddSet##SN(T* r, const T* a) {                                      \
        return coreAdd##SN(r, r, a);                                           \
    }                                                                          \
                                                                               \
    T* coreAdd##SN(T* r, const T* a, const T* b) {                             \
        for(int i = 0; i < N; i++) {                                           \
            r->data[i] = a->data[i] + b->data[i];                              \
        }                                                                      \
        return r;                                                              \
    }                                                                          \
                                                                               \
    T* coreSubSet##SN(T* r, const T* a) {                                      \
        return coreSub##SN(r, r, a);                                           \
    }                                                                          \
                                                                               \
    T* coreSub##SN(T* r, const T* a, const T* b) {                             \
        for(int i = 0; i < N; i++) {                                           \
            r->data[i] = a->data[i] - b->data[i];                              \
        }                                                                      \
        return r;                                                              \
    }                                                                          \
                                                                               \
    T* coreMulSet##SN(T* r, const T* a) {                                      \
        return coreMul##SN(r, r, a);                                           \
    }                                                                          \
                                                                               \
    T* coreMul##SN(T* r, const T* a, const T* b) {                             \
        for(int i = 0; i < N; i++) {                                           \
            r->data[i] = a->data[i] * b->data[i];                              \
        }                                                                      \
        return r;                                                              \
    }                                                                          \
                                                                               \
    T* coreDivSet##SN(T* r, const T* a) {                                      \
        return coreDiv##SN(r, r, a);                                           \
    }                                                                          \
                                                                               \
    T* coreDiv##SN(T* r, const T* a, const T* b) {                             \
        for(int i = 0; i < N; i++) {                                           \
            r->data[i] = a->data[i] / b->data[i];                              \
        }                                                                      \
        return r;                                                              \
    }                                                                          \
                                                                               \
    T* coreMulSet##SN##F(T* r, float f) {                                      \
        return coreMul##SN##F(r, r, f);                                        \
    }                                                                          \
                                                                               \
    T* coreMul##SN##F(T* r, const T* a, float f) {                             \
        for(int i = 0; i < N; i++) {                                           \
            r->data[i] = a->data[i] * f;                                       \
        }                                                                      \
        return r;                                                              \
    }                                                                          \
                                                                               \
    T* coreDivSet##SN##F(T* r, float f) {                                      \
        return coreDiv##SN##F(r, r, f);                                        \
    }                                                                          \
                                                                               \
    T* coreDiv##SN##F(T* r, const T* a, float f) {                             \
        for(int i = 0; i < N; i++) {                                           \
            r->data[i] = a->data[i] / f;                                       \
        }                                                                      \
        return r;                                                              \
    }                                                                          \
                                                                               \
    T* coreInvertSet##SN(T* r) {                                               \
        return coreInvert##SN(r, r);                                           \
    }                                                                          \
                                                                               \
    T* coreInvert##SN(T* r, const T* a) {                                      \
        for(int i = 0; i < N; i++) {                                           \
            r->data[i] = -a->data[i];                                          \
        }                                                                      \
        return r;                                                              \
    }                                                                          \
                                                                               \
    float coreDot##SN(const T* a, const T* b) {                                \
        float length = 0;                                                      \
        for(int i = 0; i < N; i++) {                                           \
            length += a->data[i] * b->data[i];                                 \
        }                                                                      \
        return length;                                                         \
    }                                                                          \
                                                                               \
    float coreSquareLength##SN(const T* a) {                                   \
        return coreDot##SN(a, a);                                              \
    }                                                                          \
                                                                               \
    float coreLength##SN(const T* a) {                                         \
        return sqrtf(coreSquareLength##SN(a));                                 \
    }                                                                          \
                                                                               \
    T* coreNormalize##SN(T* r) {                                               \
        return coreMulSet##SN##F(r, 1.0f / coreLength##SN(r));                 \
    }                                                                          \
                                                                               \
    T* coreConvertI##SN(T* r, const CN* a) {                                   \
        for(int i = 0; i < N; i++) {                                           \
            r->data[i] = (float)a->data[i];                                    \
        }                                                                      \
        return r;                                                              \
    }                                                                          \
                                                                               \
    CN* coreConvert##SN(CN* r, const T* a) {                                   \
        for(int i = 0; i < N; i++) {                                           \
            r->data[i] = (int)a->data[i];                                      \
        }                                                                      \
        return r;                                                              \
    }

VECTOR_IMPL(CoreVector2, 2, V2, F, CoreIntVector2)
VECTOR_IMPL(CoreVector3, 3, V3, F, CoreIntVector3)
VECTOR_IMPL(CoreVector4, 4, V4, F, CoreIntVector4)

void coreToStringV2(const V2* a, char* buffer, size_t n) {
    snprintf(buffer, n, "[%.3f, %.3f]", (double)a->data[0], (double)a->data[1]);
}

void coreToStringV3(const V3* a, char* buffer, size_t n) {
    snprintf(buffer, n, "[%.3f, %.3f, %.3f]", (double)a->data[0],
             (double)a->data[1], (double)a->data[2]);
}

void coreToStringV4(const V4* a, char* buffer, size_t n) {
    snprintf(buffer, n, "[%.3f, %.3f, %.3f, %.3f]", (double)a->data[0],
             (double)a->data[1], (double)a->data[2], (double)a->data[3]);
}
